#
# Copyright (c) Daniel Sheffield 2021 - 2023
#
# All rights reserved
#
# THIS SOFTWARE IS PROVIDED AS IS WITHOUT WARRANTY
from collections import (
    OrderedDict,
)
from typing import Tuple
from psycopg.sql import (
    Identifier,
    SQL,
    Literal,
    Composable,
)
from .util import get_select, get_from

def get_selectors(
    unit: str,
    product: str,
    window: SQL
) -> OrderedDict[Tuple[str, Composable]]:
    return  OrderedDict([
        ('id', Identifier('transactions', 'id')),
        ('ts_raw', SQL("""(transactions.ts AT TIME ZONE 'UTC')::timestamp without time zone""")),
        ('%d/%m/%y %_I%P', SQL("""(transactions.ts AT TIME ZONE 'UTC')::timestamp without time zone""")),
        ('code', Identifier('stores', 'code')),
        ('$/unit', SQL("""TRUNC(
    price / quantity / convert_unit(units.name, {unit}, {product}), 4
    )""").format(unit=Literal(unit), product=Literal(product))),
        ('last', SQL("""TRUNC(last_value(
    price / quantity / convert_unit(units.name, {unit}, {product})
) OVER {window}, 4)
""").format(unit=Literal(unit), product=Literal(product), window=window)),
        ('avg', SQL("""TRUNC(sum(CASE
    WHEN convert_unit(units.name, {unit}, {product}) IS NOT NULL THEN price
    ELSE NULL
END) OVER {window} / sum(
    quantity * convert_unit(units.name, {unit}, {product})
) OVER {window}, 4)
""").format(unit=Literal(unit), product=Literal(product), window=window)),
        ('min', SQL("""TRUNC(min(
    price / quantity / convert_unit(units.name, {unit}, {product})
) OVER {window}, 4)
""").format(unit=Literal(unit), product=Literal(product), window=window)),
        ('max', SQL("""TRUNC(max(
    price / quantity / convert_unit(units.name, {unit}, {product})
) OVER {window}, 4)
""").format(unit=Literal(unit), product=Literal(product), window=window)),
        ('price', SQL("""TRUNC(price, 4)""")),
        ('quantity', SQL("""TRUNC(
    quantity * convert_unit(units.name, {unit}, {product}), 4
)""").format(unit=Literal(unit), product=Literal(product))),
        ('product', Identifier('products', 'name')),
        ('category', Identifier('categories', 'name')),
        ('group', Identifier('groups', 'name')),
        ('organic', Identifier('organic')),
    ])

JOINS = OrderedDict([
    ('units', ('id', 'unit_id')),
    ('stores', ('id', 'store_id')),
    ('products', ('id', 'product_id')),
    ('categories', ('id', 'category_id')),
    ('groups', ('id', 'group_id')),
])

def get_where(product=None, category=None, group=None, organic=None, limit='90 days'):
    where = [ ]
    if product is not None:
        where.append(SQL(' ').join([
            Identifier('products', 'name'),
            SQL('='),
            Literal(product)
        ]))
    if category is not None:
        where.append(SQL(' ').join([
            Identifier('categories', 'name'),
            SQL('='),
            Literal(category)
        ]))
    if group is not None:
        where.append(SQL(' ').join([
            Identifier('groups', 'name'),
            SQL('='),
            Literal(group)
        ]))
    if organic is not None:
        where.append(SQL(' ').join([
            Identifier('organic'),
            SQL('='),
            Literal(organic),
        ]))
    if limit:
        where.append(
            SQL("{ts} at time zone 'utc' BETWEEN now() - {interval} AND now()").format(
                ts=Identifier('ts'),
                interval=SQL("{literal}::interval").format(literal=Literal(limit))
            )
        )
    return SQL('').join([
        SQL("WHERE"
            "\n      "),
        SQL("\n  AND ").join(where),
    ])

def get_historic_prices_statement(unit, sort=None, product=None, category=None, group=None, organic=None, limit='90 days'):
    window = SQL(f"""(
PARTITION BY {'organic,' if organic is not None else ''} product_id
ORDER BY convert_unit(units.name, {{unit}}, {{product}}) NULLS FIRST, ts
ROWS BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING
)""").format(unit=Literal(unit), product=Literal(product))
    organic_sort = f"{'organic,' if organic is not None else ''}"
    sort_sql = SQL('').join([
        SQL('{sort} {direction},').format(
            sort=Identifier(f'{sort}'),
            direction=SQL('DESC' if sort == 'ts' else 'ASC')
        ),
    ]) if sort is not None else SQL('')

    statement = SQL('\n').join([
        get_select(get_selectors(unit, product, window)),
        get_from("transactions", JOINS),
        get_where(product=product, category=category, group=group, organic=organic, limit=limit),
        SQL("""
ORDER BY {organic_sort} {sort} code, product, category, "group", "$/unit" ASC, ts DESC
""").format(
            sort=sort_sql,
            organic_sort=SQL(organic_sort),
       ),
    ])
    return statement
